#!/usr/bin/env node

/*
 blackcatmq
 copyright (c) 2012 Yaroslav Gaponov <yaroslav.gaponov@gmail.com>
*/

var DEBUG = true;

var util = require('util');
var net = require('net');
var uuid = require('node-uuid');
var stomp = require('./stomp');

var DELIMETER = '\0';

var sender = function(socket, data) {
    if (data) {
        if (socket && socket.writable) {
            socket.write(data.toString());
        }
    }
}

var BlackCatMQ = function(port, host, interval) {
    var self = this;
    
    if (self instanceof BlackCatMQ) {        
        self.port = port || 61613;
        self.host = host || '0.0.0.0';
        self.interval = interval || 50000;
        
        self.sockets = {};
        self.subscribes = {};
        
        self.messages = { list: {}, array: [] };
        self.ack_list = [];
        
        self.server = net.createServer({ allowHalfOpen: true }, function(socket) {            
            socket.setEncoding('utf8');
            socket.setKeepAlive(true);
            
            var remoteAddress = socket.address();
            util.log(util.format('server is connected to %s:%s', remoteAddress.address, remoteAddress.port));
            
            var data = '';
            socket.on('data', function(chunk) {                
                data += chunk.toString();
                
                var frames = data.split(DELIMETER);
                                
                if (frames.length > 1) {
                    data = frames.pop();                
                    frames.forEach(function(_frame) {
                        var frame = stomp.Frame(_frame);
                        try {
                            sender(socket, self[frame.command.toLowerCase()].call(self, socket, frame));                        
                        } catch (ex) {
                            util.log('error: ' + ex.stack);
                        }
                    });
                }   
                
            });
            
            socket.on('end', function() {
                util.log(util.format('server is disconnected from %s:%s', remoteAddress.address, remoteAddress.port));                
            });
        });        
    } else {
        return new BlackCatMQ(port, host, interval);
    }
}

BlackCatMQ.prototype.start = function() {
    var self = this;
    
    self.server.listen(self.port, self.host, function() {
        var addr = self.server.address();
        util.log(util.format("server is started on %s:%s ...", addr.address, addr.port));
        
        self.timer = setInterval(function() { self.resend(); }, self.interval);
    });    
}

BlackCatMQ.prototype.stop = function() {
    var self = this;
    
    self.server.close();
    
    clearInterval(self.timer);
    
    util.log('server is stopped');
}


BlackCatMQ.prototype.connect = function(socket, frame) {
    var self = this;
    
    var session = uuid.v4();
    self.sockets[session] = socket;
    return stomp.ServerFrame.CONNECTED(session);
}

BlackCatMQ.prototype.subscribe = function(socket, frame) {
    var self = this;
    
    if ( frame.header.ack && frame.header.ack === 'client') {
        self.ack_list.push(frame.header.session);
    }
    
    if (self.subscribes[frame.header.destination]) {
        self.subscribes[frame.header.destination].push(frame.header.session);    
    } else {
        self.subscribes[frame.header.destination] = [frame.header.session];   
    }
}

BlackCatMQ.prototype.unsubscribe = function(socket, frame) {
    var self = this;
    
    if (self.subscribes[frame.header.destination]) {
        var pos = self.subscribes[frame.header.destination].indexOf(frame.header.session);
        if (pos >= 0) {
            self.subscribes[frame.header.destination].splice(pos, 1);   
        }        
    }
}

BlackCatMQ.prototype.send = function(socket, frame) {
    var self = this;
    
    if (self.subscribes[frame.header.destination]) {
                
        var messageID = uuid.v4();
                
        if (frame.header.destination.indexOf('/queue/') === 0) {
            var session =  self.subscribes[frame.header.destination].pop();
            self.subscribes[frame.header.destination].unshift(session);
            
            
            if (self.ack_list.indexOf(session) >= 0) {
                self.messages.list[messageID] = frame;
                self.messages.array.push(messageID);
            }
            
            sender(self.sockets[session], stomp.ServerFrame.MESSAGE(frame.header.destination, messageID, frame.body));            
        } else {
            self.subscribes[frame.header.destination].forEach(function(session) {
                sender(self.sockets[session], stomp.ServerFrame.MESSAGE(frame.header.destination, messageID, frame.body));                
            });    
        }
    }
}

BlackCatMQ.prototype.ack = function(socket, frame) {
    var self = this;
    
    var messageId = frame.header['message-id'];
    
    delete self.messages.list[messageId]
    var pos = self.messages.array.indexOf(messageId);
    self.messages.array.splice(pos, 1);
    
    return stomp.ServerFrame.RECEIPT(messageId);
}

BlackCatMQ.prototype.disconnect = function(socket, frame) {
    var self = this;
    
    for(var session in self.sockets) {
        if (self.sockets[session] === socket) {
            delete self.sockets[session];
            
            for (var destination in self.subscribes) {
                var pos = self.subscribes[destination].indexOf(session);
                if (pos >= 0) {
                    self.subscribes[destination].splice(pos, 1);   
                }
            }
            
            var pos = self.ack_list.indexOf(session);
            if (pos >= 0) {
                self.ack_list.splice(pos, 1);
            }
            break;
        }        
    }    
}

BlackCatMQ.prototype.resend = function() {
    var self = this;
    
    if (self.messages.array.length) {
        var messageID = self.messages.array.shift();
        self.send(null, self.messages.list[messageID]);
        delete self.messages.list[messageID];
    }
}


var args = process.argv.splice(2);

var server;
switch(args.length) {
    case 0: server = BlackCatMQ(); break;
    case 1: server = BlackCatMQ(args[0]); break;
    case 2: server = BlackCatMQ(args[0], args[1]); break;
    default: console.log('help: BlackCatMQ {port} {host}');
}

if (server) {
    server.start();
}

process.once('uncaughtException', function(err) {
    util.debug('ERROR:' + err + err.stack);
    if (server) {
        server.stop();
    }
});

process.once('exit', function() {
    if (server) {
        server.stop();
    }
});




